/*
 * Copyright (C) by Argonne National Laboratory
 *     See COPYRIGHT in top-level directory
 */

#include <assert.h>
#include <stdbool.h>
#include <stdlib.h>
#include <limits.h>
#include <pthread.h>
#include "yaksa.h"
#include "yaksi.h"
#include "yaksu.h"
#include "yaksuri.h"
#include "yutlist.h"

static yaksuri_request_s *pending_reqs = NULL;
static pthread_mutex_t progress_mutex = PTHREAD_MUTEX_INITIALIZER;

static bool buf_is_aligned(const void *buf, yaksi_type_s * type)
{
    return !((uintptr_t) buf % type->alignment);
}

static yaksi_type_s *get_base_type(yaksi_type_s * type)
{
    int rc = YAKSA_SUCCESS;
    yaksi_type_s *ret = NULL;

    switch (type->kind) {
        case YAKSI_TYPE_KIND__BUILTIN:
            ret = type;
            break;

        case YAKSI_TYPE_KIND__CONTIG:
            ret = get_base_type(type->u.contig.child);
            break;

        case YAKSI_TYPE_KIND__RESIZED:
            ret = get_base_type(type->u.resized.child);
            break;

        case YAKSI_TYPE_KIND__HVECTOR:
            ret = get_base_type(type->u.hvector.child);
            break;

        case YAKSI_TYPE_KIND__BLKHINDX:
            ret = get_base_type(type->u.blkhindx.child);
            break;

        case YAKSI_TYPE_KIND__HINDEXED:
            ret = get_base_type(type->u.hindexed.child);
            break;

        case YAKSI_TYPE_KIND__SUBARRAY:
            ret = get_base_type(type->u.subarray.primary);
            break;

        case YAKSI_TYPE_KIND__STRUCT:
            /* structs do not have a single base type, so we treat it
             * as a collection of bytes.  this will only work for
             * contiguous + REPLACE operations, but those are the only
             * cases that we should be seeing struct types here
             * anyway. */
            assert(type->is_contig);
            rc = yaksi_type_get(YAKSA_TYPE__BYTE, &ret);
            YAKSU_ERR_CHECK(rc, fn_fail);
            break;

        default:
            ret = NULL;
            break;
    }

  fn_exit:
    return ret;
  fn_fail:
    goto fn_exit;
}

static int icopy(yaksuri_gpudriver_id_e id, const void *inbuf, void *outbuf, uintptr_t count,
                 yaksi_type_s * type, yaksi_info_s * info, yaksa_op_t op, int device)
{
    int rc = YAKSA_SUCCESS;
    yaksi_type_s *base_type = get_base_type(type);

    rc = yaksuri_global.gpudriver[id].hooks->ipack(inbuf, outbuf,
                                                   count * type->size / base_type->size,
                                                   base_type, info, op, device);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int ipack(yaksuri_gpudriver_id_e id, const void *inbuf, void *outbuf, uintptr_t count,
                 yaksi_type_s * type, yaksi_info_s * info, yaksa_op_t op, int device)
{
    int rc = YAKSA_SUCCESS;

    rc = yaksuri_global.gpudriver[id].hooks->ipack(inbuf, outbuf, count, type, info, op, device);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int iunpack(yaksuri_gpudriver_id_e id, const void *inbuf, void *outbuf, uintptr_t count,
                   yaksi_type_s * type, yaksi_info_s * info, yaksa_op_t op, int device)
{
    int rc = YAKSA_SUCCESS;

    rc = yaksuri_global.gpudriver[id].hooks->iunpack(inbuf, outbuf, count, type, info, op, device);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static bool check_p2p_comm(yaksuri_gpudriver_id_e id, int indev, int outdev)
{
    return yaksuri_global.gpudriver[id].hooks->check_p2p_comm(indev, outdev);
}

static int event_record(yaksuri_gpudriver_id_e id, int device, void **event)
{
    int rc = YAKSA_SUCCESS;

    rc = yaksuri_global.gpudriver[id].hooks->event_record(device, event);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int event_query(yaksuri_gpudriver_id_e id, void *event, int *completed)
{
    int rc = YAKSA_SUCCESS;

    rc = yaksuri_global.gpudriver[id].hooks->event_query(event, completed);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int add_dependency(yaksuri_gpudriver_id_e id, int device1, int device2)
{
    int rc = YAKSA_SUCCESS;

    rc = yaksuri_global.gpudriver[id].hooks->add_dependency(device1, device2);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int alloc_chunk(yaksuri_gpudriver_id_e id, yaksuri_request_s * reqpriv,
                       yaksuri_subreq_s * subreq, int num_tmpbufs, int *devices,
                       yaksuri_subreq_chunk_s ** chunk)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_tmpbuf_s tmpbufs[YAKSURI_SUBREQ_CHUNK_MAX_TMPBUFS];

    assert(subreq);
    assert(subreq->kind == YAKSURI_SUBREQ_KIND__MULTI_CHUNK);

    *chunk = NULL;

    for (int i = 0; i < num_tmpbufs; i++) {
        void *buf;
        if (devices[i] >= 0) {
            rc = yaksu_buffer_pool_elem_alloc(yaksuri_global.gpudriver[id].device[devices[i]],
                                              &buf);
            YAKSU_ERR_CHECK(rc, fn_fail);
        } else {
            rc = yaksu_buffer_pool_elem_alloc(yaksuri_global.gpudriver[id].host, &buf);
            YAKSU_ERR_CHECK(rc, fn_fail);
        }

        if (buf == NULL) {
            for (int j = 0; j < i; j++) {
                rc = yaksu_buffer_pool_elem_free(tmpbufs[j].pool, tmpbufs[j].buf);
                YAKSU_ERR_CHECK(rc, fn_fail);
            }
            goto fn_exit;
        } else {
            tmpbufs[i].buf = buf;
            if (devices[i] >= 0) {
                tmpbufs[i].pool = yaksuri_global.gpudriver[id].device[devices[i]];
            } else {
                tmpbufs[i].pool = yaksuri_global.gpudriver[id].host;
            }
        }
    }

    /* allocate the chunk */
    *chunk = (yaksuri_subreq_chunk_s *) malloc(sizeof(yaksuri_subreq_chunk_s));

    (*chunk)->count_offset = subreq->u.multiple.issued_count;
    uintptr_t count_per_chunk;
    count_per_chunk = YAKSURI_TMPBUF_EL_SIZE / subreq->u.multiple.type->size;
    if ((*chunk)->count_offset + count_per_chunk <= subreq->u.multiple.count) {
        (*chunk)->count = count_per_chunk;
    } else {
        (*chunk)->count = subreq->u.multiple.count - (*chunk)->count_offset;
    }

    (*chunk)->num_tmpbufs = num_tmpbufs;
    memcpy((*chunk)->tmpbufs, tmpbufs, YAKSURI_SUBREQ_CHUNK_MAX_TMPBUFS * sizeof(yaksuri_tmpbuf_s));
    (*chunk)->event = NULL;

    DL_APPEND(subreq->u.multiple.chunks, (*chunk));

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int simple_release(yaksuri_request_s * reqpriv, yaksuri_subreq_s * subreq,
                          yaksuri_subreq_chunk_s * chunk)
{
    int rc = YAKSA_SUCCESS;

    /* cleanup */
    for (int i = 0; i < chunk->num_tmpbufs; i++) {
        rc = yaksu_buffer_pool_elem_free(chunk->tmpbufs[i].pool, chunk->tmpbufs[i].buf);
        YAKSU_ERR_CHECK(rc, fn_fail);
    }

    DL_DELETE(subreq->u.multiple.chunks, chunk);
    free(chunk);

    if (subreq->u.multiple.chunks == NULL) {
        DL_DELETE(reqpriv->subreqs, subreq);
        yaksi_type_free(subreq->u.multiple.type);
        free(subreq);
    }
    if (reqpriv->subreqs == NULL) {
        HASH_DEL(pending_reqs, reqpriv);
        yaksu_atomic_decr(&reqpriv->request->cc);
    }

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int pack_d2d_p2p_acquire(yaksuri_request_s * reqpriv, yaksuri_subreq_s * subreq,
                                yaksuri_subreq_chunk_s ** chunk)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;
    yaksa_op_t op = subreq->u.multiple.op;

    assert(reqpriv->request->backend.inattr.device != reqpriv->request->backend.outattr.device);

    *chunk = NULL;

    void *base_addr = (char *) subreq->u.multiple.outbuf + subreq->u.multiple.type->true_lb;

    if (op == YAKSA_OP__REPLACE) {
        int devices[] = { reqpriv->request->backend.inattr.device };
        rc = alloc_chunk(id, reqpriv, subreq, 1, devices, chunk);
        YAKSU_ERR_CHECK(rc, fn_fail);
    } else if (buf_is_aligned(base_addr, subreq->u.multiple.type)) {
        int devices[] = { reqpriv->request->backend.inattr.device,
            reqpriv->request->backend.outattr.device
        };
        rc = alloc_chunk(id, reqpriv, subreq, 2, devices, chunk);
        YAKSU_ERR_CHECK(rc, fn_fail);
    } else {
        int devices[] = { reqpriv->request->backend.inattr.device,
            reqpriv->request->backend.outattr.device,
            reqpriv->request->backend.outattr.device
        };
        rc = alloc_chunk(id, reqpriv, subreq, 3, devices, chunk);
        YAKSU_ERR_CHECK(rc, fn_fail);
    }

    if (*chunk == NULL)
        goto fn_exit;

    void *src_d_buf, *dst_d_buf, *dst_d_buf2;
    src_d_buf = (*chunk)->tmpbufs[0].buf;
    dst_d_buf = (*chunk)->tmpbufs[1].buf;
    dst_d_buf2 = (*chunk)->tmpbufs[2].buf;

    const char *sbuf;
    char *dbuf;
    sbuf = (const char *) subreq->u.multiple.inbuf +
        (*chunk)->count_offset * subreq->u.multiple.type->extent;
    dbuf =
        (char *) subreq->u.multiple.outbuf + (*chunk)->count_offset * subreq->u.multiple.type->size;

    rc = ipack(id, sbuf, src_d_buf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
               YAKSA_OP__REPLACE, reqpriv->request->backend.inattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    if (op == YAKSA_OP__REPLACE) {
        rc = icopy(id, src_d_buf, dbuf, (*chunk)->count, subreq->u.multiple.type,
                   reqpriv->info, YAKSA_OP__REPLACE, reqpriv->request->backend.inattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);

        rc = event_record(id, reqpriv->request->backend.inattr.device, &(*chunk)->event);
        YAKSU_ERR_CHECK(rc, fn_fail);
    } else if (buf_is_aligned(base_addr, subreq->u.multiple.type)) {
        rc = icopy(id, src_d_buf, dst_d_buf, (*chunk)->count, subreq->u.multiple.type,
                   reqpriv->info, YAKSA_OP__REPLACE, reqpriv->request->backend.inattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);

        rc = add_dependency(id, reqpriv->request->backend.inattr.device,
                            reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);

        rc = icopy(id, dst_d_buf, dbuf, (*chunk)->count, subreq->u.multiple.type,
                   reqpriv->info, op, reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);

        rc = event_record(id, reqpriv->request->backend.outattr.device, &(*chunk)->event);
        YAKSU_ERR_CHECK(rc, fn_fail);
    } else {
        rc = icopy(id, src_d_buf, dst_d_buf, (*chunk)->count, subreq->u.multiple.type,
                   reqpriv->info, YAKSA_OP__REPLACE, reqpriv->request->backend.inattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);

        rc = icopy(id, dbuf, dst_d_buf2, (*chunk)->count, subreq->u.multiple.type,
                   reqpriv->info, YAKSA_OP__REPLACE, reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);

        rc = add_dependency(id, reqpriv->request->backend.inattr.device,
                            reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);

        rc = icopy(id, dst_d_buf, dst_d_buf2, (*chunk)->count, subreq->u.multiple.type,
                   reqpriv->info, op, reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);

        rc = icopy(id, dst_d_buf2, dbuf, (*chunk)->count, subreq->u.multiple.type,
                   reqpriv->info, YAKSA_OP__REPLACE, reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);

        rc = event_record(id, reqpriv->request->backend.outattr.device, &(*chunk)->event);
        YAKSU_ERR_CHECK(rc, fn_fail);
    }

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int pack_d2d_nop2p_acquire(yaksuri_request_s * reqpriv, yaksuri_subreq_s * subreq,
                                  yaksuri_subreq_chunk_s ** chunk)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;
    yaksa_op_t op = subreq->u.multiple.op;

    assert(reqpriv->request->backend.inattr.device != reqpriv->request->backend.outattr.device);

    *chunk = NULL;

    void *base_addr = (char *) subreq->u.multiple.outbuf + subreq->u.multiple.type->true_lb;

    if (op == YAKSA_OP__REPLACE) {
        int devices[] = { reqpriv->request->backend.inattr.device, -1 };
        rc = alloc_chunk(id, reqpriv, subreq, 2, devices, chunk);
        YAKSU_ERR_CHECK(rc, fn_fail);
    } else if (buf_is_aligned(base_addr, subreq->u.multiple.type)) {
        int devices[] = { reqpriv->request->backend.inattr.device, -1,
            reqpriv->request->backend.outattr.device
        };
        rc = alloc_chunk(id, reqpriv, subreq, 3, devices, chunk);
        YAKSU_ERR_CHECK(rc, fn_fail);
    } else {
        int devices[] = { reqpriv->request->backend.inattr.device, -1,
            reqpriv->request->backend.outattr.device,
            reqpriv->request->backend.outattr.device
        };
        rc = alloc_chunk(id, reqpriv, subreq, 4, devices, chunk);
        YAKSU_ERR_CHECK(rc, fn_fail);
    }

    if (*chunk == NULL)
        goto fn_exit;

    void *src_d_buf, *rh_buf, *dst_d_buf, *dst_d_buf2;
    src_d_buf = (*chunk)->tmpbufs[0].buf;
    rh_buf = (*chunk)->tmpbufs[1].buf;
    dst_d_buf = (*chunk)->tmpbufs[2].buf;
    dst_d_buf2 = (*chunk)->tmpbufs[3].buf;

    const char *sbuf;
    char *dbuf;
    sbuf = (const char *) subreq->u.multiple.inbuf +
        (*chunk)->count_offset * subreq->u.multiple.type->extent;
    dbuf =
        (char *) subreq->u.multiple.outbuf + (*chunk)->count_offset * subreq->u.multiple.type->size;

    rc = ipack(id, sbuf, src_d_buf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
               YAKSA_OP__REPLACE, reqpriv->request->backend.inattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = icopy(id, src_d_buf, rh_buf, (*chunk)->count, subreq->u.multiple.type,
               reqpriv->info, YAKSA_OP__REPLACE, reqpriv->request->backend.inattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = add_dependency(id, reqpriv->request->backend.inattr.device,
                        reqpriv->request->backend.outattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    if (op == YAKSA_OP__REPLACE) {
        rc = icopy(id, rh_buf, dbuf, (*chunk)->count, subreq->u.multiple.type,
                   reqpriv->info, YAKSA_OP__REPLACE, reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);
    } else if (buf_is_aligned(base_addr, subreq->u.multiple.type)) {
        rc = icopy(id, rh_buf, dst_d_buf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
                   YAKSA_OP__REPLACE, reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);

        rc = icopy(id, dst_d_buf, dbuf, (*chunk)->count, subreq->u.multiple.type,
                   reqpriv->info, op, reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);
    } else {
        rc = icopy(id, rh_buf, dst_d_buf, (*chunk)->count, subreq->u.multiple.type,
                   reqpriv->info, YAKSA_OP__REPLACE, reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);

        rc = icopy(id, dbuf, dst_d_buf2, (*chunk)->count, subreq->u.multiple.type,
                   reqpriv->info, YAKSA_OP__REPLACE, reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);

        rc = icopy(id, dst_d_buf, dst_d_buf2, (*chunk)->count, subreq->u.multiple.type,
                   reqpriv->info, op, reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);

        rc = icopy(id, dst_d_buf2, dbuf, (*chunk)->count, subreq->u.multiple.type,
                   reqpriv->info, YAKSA_OP__REPLACE, reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);
    }

    rc = event_record(id, reqpriv->request->backend.outattr.device, &(*chunk)->event);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int pack_d2d_unaligned_acquire(yaksuri_request_s * reqpriv, yaksuri_subreq_s * subreq,
                                      yaksuri_subreq_chunk_s ** chunk)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;
    yaksa_op_t op = subreq->u.multiple.op;

    *chunk = NULL;

    int devices[] = { reqpriv->request->backend.inattr.device };
    rc = alloc_chunk(id, reqpriv, subreq, 1, devices, chunk);
    YAKSU_ERR_CHECK(rc, fn_fail);

    if (*chunk == NULL)
        goto fn_exit;

    void *d_buf;
    d_buf = (*chunk)->tmpbufs[0].buf;

    const char *sbuf;
    char *dbuf;
    sbuf = (const char *) subreq->u.multiple.inbuf +
        (*chunk)->count_offset * subreq->u.multiple.type->extent;
    dbuf =
        (char *) subreq->u.multiple.outbuf + (*chunk)->count_offset * subreq->u.multiple.type->size;

    rc = icopy(id, dbuf, d_buf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
               YAKSA_OP__REPLACE, reqpriv->request->backend.inattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = ipack(id, sbuf, d_buf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
               op, reqpriv->request->backend.inattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = icopy(id, d_buf, dbuf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
               YAKSA_OP__REPLACE, reqpriv->request->backend.inattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = event_record(id, reqpriv->request->backend.inattr.device, &(*chunk)->event);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int pack_d2rh_acquire(yaksuri_request_s * reqpriv, yaksuri_subreq_s * subreq,
                             yaksuri_subreq_chunk_s ** chunk)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;
    yaksa_op_t op = subreq->u.multiple.op;

    *chunk = NULL;

    if (op == YAKSA_OP__REPLACE) {
        int devices[] = { reqpriv->request->backend.inattr.device };
        rc = alloc_chunk(id, reqpriv, subreq, 1, devices, chunk);
        YAKSU_ERR_CHECK(rc, fn_fail);
    } else {
        int devices[] = { reqpriv->request->backend.inattr.device, -1 };
        rc = alloc_chunk(id, reqpriv, subreq, 2, devices, chunk);
        YAKSU_ERR_CHECK(rc, fn_fail);
    }

    if (*chunk == NULL)
        goto fn_exit;

    void *d_buf, *rh_buf;
    d_buf = (*chunk)->tmpbufs[0].buf;
    rh_buf = (*chunk)->tmpbufs[1].buf;

    const char *sbuf;
    char *dbuf;
    sbuf = (const char *) subreq->u.multiple.inbuf +
        (*chunk)->count_offset * subreq->u.multiple.type->extent;
    dbuf = (char *) subreq->u.multiple.outbuf + (*chunk)->count_offset *
        subreq->u.multiple.type->size;

    rc = ipack(id, sbuf, d_buf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
               YAKSA_OP__REPLACE, reqpriv->request->backend.inattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    if (op == YAKSA_OP__REPLACE) {
        rc = icopy(id, d_buf, dbuf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
                   YAKSA_OP__REPLACE, reqpriv->request->backend.inattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);
    } else {
        rc = icopy(id, d_buf, rh_buf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
                   YAKSA_OP__REPLACE, reqpriv->request->backend.inattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);
    }

    rc = event_record(id, reqpriv->request->backend.inattr.device, &(*chunk)->event);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int pack_d2rh_release(yaksuri_request_s * reqpriv, yaksuri_subreq_s * subreq,
                             yaksuri_subreq_chunk_s * chunk)
{
    int rc = YAKSA_SUCCESS;
    yaksa_op_t op = subreq->u.multiple.op;
    yaksi_type_s *type = subreq->u.multiple.type;

    if (op != YAKSA_OP__REPLACE) {
        yaksi_type_s *base_type = get_base_type(type);
        char *dbuf = (char *) subreq->u.multiple.outbuf + chunk->count_offset *
            subreq->u.multiple.type->size;

        rc = yaksuri_seq_ipack(chunk->tmpbufs[1].buf, dbuf,
                               chunk->count * type->size / base_type->size, base_type,
                               reqpriv->info, subreq->u.multiple.op);
        YAKSU_ERR_CHECK(rc, fn_fail);
    }

    rc = simple_release(reqpriv, subreq, chunk);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int pack_d2urh_acquire(yaksuri_request_s * reqpriv, yaksuri_subreq_s * subreq,
                              yaksuri_subreq_chunk_s ** chunk)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;

    *chunk = NULL;

    int devices[] = { reqpriv->request->backend.inattr.device, -1 };
    rc = alloc_chunk(id, reqpriv, subreq, 2, devices, chunk);
    YAKSU_ERR_CHECK(rc, fn_fail);

    if (*chunk == NULL)
        goto fn_exit;

    void *d_buf, *rh_buf;
    d_buf = (*chunk)->tmpbufs[0].buf;
    rh_buf = (*chunk)->tmpbufs[1].buf;

    const char *sbuf;
    sbuf = (const char *) subreq->u.multiple.inbuf +
        (*chunk)->count_offset * subreq->u.multiple.type->extent;

    rc = ipack(id, sbuf, d_buf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
               YAKSA_OP__REPLACE, reqpriv->request->backend.inattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = icopy(id, d_buf, rh_buf, (*chunk)->count, subreq->u.multiple.type,
               reqpriv->info, YAKSA_OP__REPLACE, reqpriv->request->backend.inattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = event_record(id, reqpriv->request->backend.inattr.device, &(*chunk)->event);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int pack_d2urh_release(yaksuri_request_s * reqpriv, yaksuri_subreq_s * subreq,
                              yaksuri_subreq_chunk_s * chunk)
{
    int rc = YAKSA_SUCCESS;
    char *dbuf = (char *) subreq->u.multiple.outbuf + chunk->count_offset *
        subreq->u.multiple.type->size;
    yaksi_type_s *type = subreq->u.multiple.type;
    yaksi_type_s *base_type = get_base_type(type);

    rc = yaksuri_seq_ipack(chunk->tmpbufs[1].buf, dbuf,
                           chunk->count * type->size / base_type->size, base_type,
                           reqpriv->info, subreq->u.multiple.op);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = simple_release(reqpriv, subreq, chunk);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int pack_h2d_acquire(yaksuri_request_s * reqpriv, yaksuri_subreq_s * subreq,
                            yaksuri_subreq_chunk_s ** chunk)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;
    yaksa_op_t op = subreq->u.multiple.op;

    *chunk = NULL;

    void *base_addr = (char *) subreq->u.multiple.outbuf + subreq->u.multiple.type->true_lb;

    if (op == YAKSA_OP__REPLACE) {
        int devices[] = { -1 };
        rc = alloc_chunk(id, reqpriv, subreq, 1, devices, chunk);
        YAKSU_ERR_CHECK(rc, fn_fail);
    } else if (buf_is_aligned(base_addr, subreq->u.multiple.type)) {
        int devices[] = { -1, reqpriv->request->backend.outattr.device };
        rc = alloc_chunk(id, reqpriv, subreq, 2, devices, chunk);
        YAKSU_ERR_CHECK(rc, fn_fail);
    } else {
        int devices[] = { -1, reqpriv->request->backend.outattr.device,
            reqpriv->request->backend.outattr.device
        };
        rc = alloc_chunk(id, reqpriv, subreq, 3, devices, chunk);
        YAKSU_ERR_CHECK(rc, fn_fail);
    }

    if (*chunk == NULL)
        goto fn_exit;

    void *rh_buf, *d_buf, *d_buf2;
    rh_buf = (*chunk)->tmpbufs[0].buf;
    d_buf = (*chunk)->tmpbufs[1].buf;
    d_buf2 = (*chunk)->tmpbufs[2].buf;

    const char *sbuf;
    char *dbuf;
    sbuf = (const char *) subreq->u.multiple.inbuf +
        (*chunk)->count_offset * subreq->u.multiple.type->extent;
    dbuf = (char *) subreq->u.multiple.outbuf + (*chunk)->count_offset *
        subreq->u.multiple.type->size;

    rc = yaksuri_seq_ipack(sbuf, rh_buf, (*chunk)->count, subreq->u.multiple.type,
                           reqpriv->info, YAKSA_OP__REPLACE);
    YAKSU_ERR_CHECK(rc, fn_fail);

    if (op == YAKSA_OP__REPLACE) {
        rc = icopy(id, rh_buf, dbuf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
                   YAKSA_OP__REPLACE, reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);
    } else if (buf_is_aligned(base_addr, subreq->u.multiple.type)) {
        rc = icopy(id, rh_buf, d_buf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
                   YAKSA_OP__REPLACE, reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);

        rc = icopy(id, d_buf, dbuf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
                   op, reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);
    } else {
        rc = icopy(id, rh_buf, d_buf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
                   YAKSA_OP__REPLACE, reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);

        rc = icopy(id, dbuf, d_buf2, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
                   YAKSA_OP__REPLACE, reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);

        rc = icopy(id, d_buf, d_buf2, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
                   op, reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);

        rc = icopy(id, d_buf2, dbuf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
                   YAKSA_OP__REPLACE, reqpriv->request->backend.outattr.device);
        YAKSU_ERR_CHECK(rc, fn_fail);
    }

    rc = event_record(id, reqpriv->request->backend.outattr.device, &(*chunk)->event);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int unpack_d2d_p2p_acquire(yaksuri_request_s * reqpriv, yaksuri_subreq_s * subreq,
                                  yaksuri_subreq_chunk_s ** chunk)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;

    assert(reqpriv->request->backend.inattr.device != reqpriv->request->backend.outattr.device);

    *chunk = NULL;

    int devices[] = { reqpriv->request->backend.outattr.device };
    rc = alloc_chunk(id, reqpriv, subreq, 1, devices, chunk);
    YAKSU_ERR_CHECK(rc, fn_fail);

    if (*chunk == NULL)
        goto fn_exit;

    void *d_buf;
    d_buf = (*chunk)->tmpbufs[0].buf;

    const char *sbuf;
    char *dbuf;
    sbuf = (const char *) subreq->u.multiple.inbuf +
        (*chunk)->count_offset * subreq->u.multiple.type->size;
    dbuf = (char *) subreq->u.multiple.outbuf +
        (*chunk)->count_offset * subreq->u.multiple.type->extent;

    rc = icopy(id, sbuf, d_buf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
               YAKSA_OP__REPLACE, reqpriv->request->backend.inattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = add_dependency(id, reqpriv->request->backend.inattr.device,
                        reqpriv->request->backend.outattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = iunpack(id, d_buf, dbuf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
                 subreq->u.multiple.op, reqpriv->request->backend.outattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = event_record(id, reqpriv->request->backend.outattr.device, &(*chunk)->event);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int unpack_d2d_nop2p_acquire(yaksuri_request_s * reqpriv, yaksuri_subreq_s * subreq,
                                    yaksuri_subreq_chunk_s ** chunk)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;

    assert(reqpriv->request->backend.inattr.device != reqpriv->request->backend.outattr.device);

    *chunk = NULL;

    int devices[] = { reqpriv->request->backend.outattr.device, -1 };
    rc = alloc_chunk(id, reqpriv, subreq, 2, devices, chunk);
    YAKSU_ERR_CHECK(rc, fn_fail);

    if (*chunk == NULL)
        goto fn_exit;

    void *d_buf, *rh_buf;
    d_buf = (*chunk)->tmpbufs[0].buf;
    rh_buf = (*chunk)->tmpbufs[1].buf;

    const char *sbuf;
    char *dbuf;
    sbuf = (const char *) subreq->u.multiple.inbuf +
        (*chunk)->count_offset * subreq->u.multiple.type->size;
    dbuf = (char *) subreq->u.multiple.outbuf +
        (*chunk)->count_offset * subreq->u.multiple.type->extent;

    rc = icopy(id, sbuf, rh_buf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
               YAKSA_OP__REPLACE, reqpriv->request->backend.inattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = add_dependency(id, reqpriv->request->backend.inattr.device,
                        reqpriv->request->backend.outattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = icopy(id, rh_buf, d_buf, (*chunk)->count, subreq->u.multiple.type,
               reqpriv->info, YAKSA_OP__REPLACE, reqpriv->request->backend.outattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = iunpack(id, d_buf, dbuf, (*chunk)->count, subreq->u.multiple.type,
                 reqpriv->info, subreq->u.multiple.op, reqpriv->request->backend.outattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = event_record(id, reqpriv->request->backend.outattr.device, &(*chunk)->event);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int unpack_d2d_unaligned_acquire(yaksuri_request_s * reqpriv, yaksuri_subreq_s * subreq,
                                        yaksuri_subreq_chunk_s ** chunk)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;

    *chunk = NULL;

    int devices[] = { reqpriv->request->backend.inattr.device };
    rc = alloc_chunk(id, reqpriv, subreq, 1, devices, chunk);
    YAKSU_ERR_CHECK(rc, fn_fail);

    if (*chunk == NULL)
        goto fn_exit;

    void *d_buf;
    d_buf = (*chunk)->tmpbufs[0].buf;

    const char *sbuf;
    char *dbuf;
    sbuf = (const char *) subreq->u.multiple.inbuf +
        (*chunk)->count_offset * subreq->u.multiple.type->size;
    dbuf = (char *) subreq->u.multiple.outbuf +
        (*chunk)->count_offset * subreq->u.multiple.type->extent;

    rc = icopy(id, sbuf, d_buf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
               YAKSA_OP__REPLACE, reqpriv->request->backend.inattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = iunpack(id, d_buf, dbuf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
                 subreq->u.multiple.op, reqpriv->request->backend.inattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = event_record(id, reqpriv->request->backend.inattr.device, &(*chunk)->event);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int unpack_rh2d_acquire(yaksuri_request_s * reqpriv, yaksuri_subreq_s * subreq,
                               yaksuri_subreq_chunk_s ** chunk)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;

    *chunk = NULL;

    int devices[] = { reqpriv->request->backend.outattr.device };
    rc = alloc_chunk(id, reqpriv, subreq, 1, devices, chunk);
    YAKSU_ERR_CHECK(rc, fn_fail);

    if (*chunk == NULL)
        goto fn_exit;

    void *d_buf;
    d_buf = (*chunk)->tmpbufs[0].buf;

    const char *sbuf;
    char *dbuf;
    sbuf = (const char *) subreq->u.multiple.inbuf +
        (*chunk)->count_offset * subreq->u.multiple.type->size;
    dbuf = (char *) subreq->u.multiple.outbuf +
        (*chunk)->count_offset * subreq->u.multiple.type->extent;

    rc = icopy(id, sbuf, d_buf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
               YAKSA_OP__REPLACE, reqpriv->request->backend.outattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = iunpack(id, d_buf, dbuf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
                 subreq->u.multiple.op, reqpriv->request->backend.outattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = event_record(id, reqpriv->request->backend.outattr.device, &(*chunk)->event);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int unpack_urh2d_acquire(yaksuri_request_s * reqpriv, yaksuri_subreq_s * subreq,
                                yaksuri_subreq_chunk_s ** chunk)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;

    *chunk = NULL;

    int devices[] = { reqpriv->request->backend.outattr.device, -1 };
    rc = alloc_chunk(id, reqpriv, subreq, 2, devices, chunk);
    YAKSU_ERR_CHECK(rc, fn_fail);

    if (*chunk == NULL)
        goto fn_exit;

    void *d_buf, *rh_buf;
    d_buf = (*chunk)->tmpbufs[0].buf;
    rh_buf = (*chunk)->tmpbufs[1].buf;

    const char *sbuf;
    char *dbuf;
    sbuf = (const char *) subreq->u.multiple.inbuf +
        (*chunk)->count_offset * subreq->u.multiple.type->size;
    dbuf = (char *) subreq->u.multiple.outbuf +
        (*chunk)->count_offset * subreq->u.multiple.type->extent;

    yaksi_type_s *byte_type;
    rc = yaksi_type_get(YAKSA_TYPE__BYTE, &byte_type);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = yaksuri_seq_ipack(sbuf, rh_buf, (*chunk)->count * subreq->u.multiple.type->size,
                           byte_type, reqpriv->info, YAKSA_OP__REPLACE);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = icopy(id, rh_buf, d_buf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
               YAKSA_OP__REPLACE, reqpriv->request->backend.outattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = iunpack(id, d_buf, dbuf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
                 subreq->u.multiple.op, reqpriv->request->backend.outattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = event_record(id, reqpriv->request->backend.outattr.device, &(*chunk)->event);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int unpack_d2h_acquire(yaksuri_request_s * reqpriv, yaksuri_subreq_s * subreq,
                              yaksuri_subreq_chunk_s ** chunk)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;

    *chunk = NULL;

    int devices[] = { -1 };
    rc = alloc_chunk(id, reqpriv, subreq, 1, devices, chunk);
    YAKSU_ERR_CHECK(rc, fn_fail);

    if (*chunk == NULL)
        goto fn_exit;

    void *rh_buf;
    rh_buf = (*chunk)->tmpbufs[0].buf;

    const char *sbuf;
    sbuf = (const char *) subreq->u.multiple.inbuf +
        (*chunk)->count_offset * subreq->u.multiple.type->size;

    rc = icopy(id, sbuf, rh_buf, (*chunk)->count, subreq->u.multiple.type, reqpriv->info,
               YAKSA_OP__REPLACE, reqpriv->request->backend.inattr.device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = event_record(id, reqpriv->request->backend.inattr.device, &(*chunk)->event);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int unpack_d2h_release(yaksuri_request_s * reqpriv, yaksuri_subreq_s * subreq,
                              yaksuri_subreq_chunk_s * chunk)
{
    int rc = YAKSA_SUCCESS;

    char *dbuf;
    dbuf = (char *) subreq->u.multiple.outbuf + chunk->count_offset *
        subreq->u.multiple.type->extent;
    rc = yaksuri_seq_iunpack(chunk->tmpbufs[0].buf, dbuf, chunk->count, subreq->u.multiple.type,
                             reqpriv->info, subreq->u.multiple.op);
    YAKSU_ERR_CHECK(rc, fn_fail);

    rc = simple_release(reqpriv, subreq, chunk);
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int set_multichunk_subreq(const void *inbuf, void *outbuf, uintptr_t count,
                                 yaksi_type_s * type, yaksa_op_t op, yaksuri_subreq_s ** subreq_ptr)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_subreq_s *subreq = (yaksuri_subreq_s *) malloc(sizeof(yaksuri_subreq_s));
    *subreq_ptr = subreq;

    /* we can only take on types where at least one count of the type
     * fits into our temporary buffers. */
    if (type->size > YAKSURI_TMPBUF_EL_SIZE) {
        return YAKSA_ERR__NOT_SUPPORTED;
    }

    subreq->kind = YAKSURI_SUBREQ_KIND__MULTI_CHUNK;
    subreq->u.multiple.inbuf = inbuf;
    subreq->u.multiple.outbuf = outbuf;
    subreq->u.multiple.count = count;
    subreq->u.multiple.type = type;
    subreq->u.multiple.op = op;
    subreq->u.multiple.issued_count = 0;
    subreq->u.multiple.chunks = NULL;

    yaksu_atomic_incr(&type->refcount);
    return rc;
}

/* Fast-path pack subroutine */
static int singlechunk_pack(yaksuri_gpudriver_id_e id, int device, const void *inbuf, void *outbuf,
                            uintptr_t count, yaksi_type_s * type, yaksi_info_s * info,
                            yaksa_op_t op, yaksi_request_s * request,
                            yaksuri_subreq_s ** subreq_ptr)
{
    int rc = YAKSA_SUCCESS;

    rc = ipack(id, inbuf, outbuf, count, type, info, op, device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    /* Try to complete immediately for blocking request to avoid overhead
     * caused by subreq enqueue and event management */
    if (request->is_blocking && yaksuri_global.gpudriver[id].hooks->synchronize) {
        yaksuri_global.gpudriver[id].hooks->synchronize(device);
    } else {
        yaksuri_subreq_s *subreq = (yaksuri_subreq_s *) malloc(sizeof(yaksuri_subreq_s));
        subreq->kind = YAKSURI_SUBREQ_KIND__SINGLE_CHUNK;

        rc = event_record(id, device, &subreq->u.single.event);
        *subreq_ptr = subreq;
    }

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

/* Fast-path unpack subroutine */
static int singlechunk_unpack(yaksuri_gpudriver_id_e id, int device, const void *inbuf,
                              void *outbuf, uintptr_t count, yaksi_type_s * type,
                              yaksi_info_s * info, yaksa_op_t op, yaksi_request_s * request,
                              yaksuri_subreq_s ** subreq_ptr)
{
    int rc = YAKSA_SUCCESS;

    rc = iunpack(id, inbuf, outbuf, count, type, info, op, device);
    YAKSU_ERR_CHECK(rc, fn_fail);

    /* Try to complete immediately for blocking request to avoid overhead
     * caused by subreq enqueue and event management */
    if (request->is_blocking && yaksuri_global.gpudriver[id].hooks->synchronize) {
        yaksuri_global.gpudriver[id].hooks->synchronize(device);
    } else {
        yaksuri_subreq_s *subreq = (yaksuri_subreq_s *) malloc(sizeof(yaksuri_subreq_s));
        subreq->kind = YAKSURI_SUBREQ_KIND__SINGLE_CHUNK;

        rc = event_record(id, device, &subreq->u.single.event);
        *subreq_ptr = subreq;
    }

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

/* Subroutines to setup subreq for pack with different buffer types */
static int set_subreq_pack_d2d(const void *inbuf, void *outbuf, uintptr_t count,
                               yaksi_type_s * type, yaksi_info_s * info, yaksa_op_t op,
                               yaksi_request_s * request, yaksuri_request_s * reqpriv,
                               yaksuri_subreq_s ** subreq_ptr)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;
    bool aligned = buf_is_aligned(outbuf, type);

    /* Fast path for REPLACE with aligned outbuf on the same device or different
     * devices with P2P support. Note that IPC mapping requires P2P support, thus
     * we don't handle the IPC case explicitly. */
    if (op == YAKSA_OP__REPLACE &&
        (request->backend.inattr.device == request->backend.outattr.device ||
         check_p2p_comm(id, reqpriv->request->backend.inattr.device,
                        reqpriv->request->backend.outattr.device)) && aligned) {
        rc = singlechunk_pack(id, request->backend.inattr.device, inbuf, outbuf, count,
                              type, info, op, request, subreq_ptr);
    }
    /* Fast path for other reduce operations with aligned buffer on the same device */
    else if (request->backend.inattr.device == request->backend.outattr.device && aligned) {
        rc = singlechunk_pack(id, request->backend.inattr.device, inbuf, outbuf, count,
                              type, info, op, request, subreq_ptr);
    }
    /* Slow paths */
    else if (request->backend.inattr.device == request->backend.outattr.device) {
        rc = set_multichunk_subreq(inbuf, outbuf, count, type, op, subreq_ptr);
        YAKSU_ERR_CHECK(rc, fn_fail);

        (*subreq_ptr)->u.multiple.acquire = pack_d2d_unaligned_acquire;
        (*subreq_ptr)->u.multiple.release = simple_release;
    } else {
        bool is_enabled = check_p2p_comm(id, reqpriv->request->backend.inattr.device,
                                         reqpriv->request->backend.outattr.device);
        rc = set_multichunk_subreq(inbuf, outbuf, count, type, op, subreq_ptr);
        YAKSU_ERR_CHECK(rc, fn_fail);

        if (is_enabled) {
            (*subreq_ptr)->u.multiple.acquire = pack_d2d_p2p_acquire;
        } else {
            (*subreq_ptr)->u.multiple.acquire = pack_d2d_nop2p_acquire;
        }
        (*subreq_ptr)->u.multiple.release = simple_release;
    }

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int set_subreq_pack_d2m(const void *inbuf, void *outbuf, uintptr_t count,
                               yaksi_type_s * type, yaksi_info_s * info, yaksa_op_t op,
                               yaksi_request_s * request, yaksuri_request_s * reqpriv,
                               yaksuri_subreq_s ** subreq_ptr)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;

    /* Fast path for all reduce operations with aligned outbuf on host or the same device */
    if ((request->backend.outattr.device == -1 ||
         request->backend.inattr.device == request->backend.outattr.device) &&
        buf_is_aligned(outbuf, type)) {
        rc = singlechunk_pack(id, request->backend.inattr.device, inbuf, outbuf, count, type, info,
                              op, request, subreq_ptr);
    }
    /* Slow paths */
    else if (request->backend.outattr.device == -1 ||
             request->backend.inattr.device == request->backend.outattr.device) {
        rc = set_multichunk_subreq(inbuf, outbuf, count, type, op, subreq_ptr);
        YAKSU_ERR_CHECK(rc, fn_fail);

        (*subreq_ptr)->u.multiple.acquire = pack_d2d_unaligned_acquire;
        (*subreq_ptr)->u.multiple.release = simple_release;
    } else {
        rc = set_multichunk_subreq(inbuf, outbuf, count, type, op, subreq_ptr);
        YAKSU_ERR_CHECK(rc, fn_fail);

        (*subreq_ptr)->u.multiple.acquire = pack_d2urh_acquire;
        (*subreq_ptr)->u.multiple.release = pack_d2urh_release;
    }

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int set_subreq_pack_d2rh(const void *inbuf, void *outbuf, uintptr_t count,
                                yaksi_type_s * type, yaksi_info_s * info, yaksa_op_t op,
                                yaksi_request_s * request, yaksuri_request_s * reqpriv,
                                yaksuri_subreq_s ** subreq_ptr)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;
    uintptr_t threshold = yaksuri_global.gpudriver[id].hooks->get_iov_pack_threshold(info);

    /* Fast path for REPLACE with contig type or noncontig type with large contig chunk */
    if (op == YAKSA_OP__REPLACE && (type->is_contig || type->size / type->num_contig >= threshold)) {
        rc = singlechunk_pack(id, request->backend.inattr.device, inbuf, outbuf, count,
                              type, info, op, request, subreq_ptr);
    }
    /* Slow path */
    else {
        rc = set_multichunk_subreq(inbuf, outbuf, count, type, op, subreq_ptr);
        YAKSU_ERR_CHECK(rc, fn_fail);

        (*subreq_ptr)->u.multiple.acquire = pack_d2rh_acquire;
        (*subreq_ptr)->u.multiple.release = pack_d2rh_release;
    }

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int set_subreq_pack_from_device(const void *inbuf, void *outbuf, uintptr_t count,
                                       yaksi_type_s * type, yaksi_info_s * info, yaksa_op_t op,
                                       yaksi_request_s * request, yaksuri_request_s * reqpriv,
                                       yaksuri_subreq_s ** subreq_ptr)
{
    int rc = YAKSA_SUCCESS;

    switch (request->backend.outattr.type) {
        case YAKSUR_PTR_TYPE__GPU:
            rc = set_subreq_pack_d2d(inbuf, outbuf, count, type, info, op, request, reqpriv,
                                     subreq_ptr);
            break;
        case YAKSUR_PTR_TYPE__MANAGED:
            rc = set_subreq_pack_d2m(inbuf, outbuf, count, type, info, op, request, reqpriv,
                                     subreq_ptr);
            break;
        case YAKSUR_PTR_TYPE__REGISTERED_HOST:
            rc = set_subreq_pack_d2rh(inbuf, outbuf, count, type, info, op, request, reqpriv,
                                      subreq_ptr);
            break;
        case YAKSUR_PTR_TYPE__UNREGISTERED_HOST:
        default:
            rc = set_multichunk_subreq(inbuf, outbuf, count, type, op, subreq_ptr);
            YAKSU_ERR_CHECK(rc, fn_fail);

            (*subreq_ptr)->u.multiple.acquire = pack_d2urh_acquire;
            (*subreq_ptr)->u.multiple.release = pack_d2urh_release;
    }

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int set_subreq_pack_from_managed(const void *inbuf, void *outbuf, uintptr_t count,
                                        yaksi_type_s * type, yaksi_info_s * info, yaksa_op_t op,
                                        yaksi_request_s * request, yaksuri_request_s * reqpriv,
                                        yaksuri_subreq_s ** subreq_ptr)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;

    /* Fast path for all reduce operations with aligned outbuf from host or the same device. */
    if ((request->backend.inattr.device == -1 ||
         request->backend.inattr.device == request->backend.outattr.device) &&
        buf_is_aligned(outbuf, type)) {
        rc = singlechunk_pack(id, request->backend.outattr.device, inbuf, outbuf, count, type, info,
                              op, request, subreq_ptr);
    }
    /* Slow path */
    else {
        rc = set_multichunk_subreq(inbuf, outbuf, count, type, op, subreq_ptr);
        YAKSU_ERR_CHECK(rc, fn_fail);

        (*subreq_ptr)->u.multiple.acquire = pack_h2d_acquire;
        (*subreq_ptr)->u.multiple.release = simple_release;
    }

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int set_subreq_pack_from_rhost(const void *inbuf, void *outbuf, uintptr_t count,
                                      yaksi_type_s * type, yaksi_info_s * info, yaksa_op_t op,
                                      yaksi_request_s * request, yaksuri_request_s * reqpriv,
                                      yaksuri_subreq_s ** subreq_ptr)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;
    uintptr_t threshold = yaksuri_global.gpudriver[id].hooks->get_iov_pack_threshold(info);

    /* Fast path for REPLACE with contig type or noncontig type with large contig chunk */
    if (op == YAKSA_OP__REPLACE && (type->is_contig || type->size / type->num_contig >= threshold)) {
        rc = singlechunk_pack(id, request->backend.outattr.device, inbuf, outbuf, count,
                              type, info, op, request, subreq_ptr);
    }
    /* Slow path */
    else {
        rc = set_multichunk_subreq(inbuf, outbuf, count, type, op, subreq_ptr);
        YAKSU_ERR_CHECK(rc, fn_fail);

        (*subreq_ptr)->u.multiple.acquire = pack_h2d_acquire;
        (*subreq_ptr)->u.multiple.release = simple_release;
    }

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

/* Subroutines to setup subreq for unpack with different buffer types */
static int set_subreq_unpack_d2d(const void *inbuf, void *outbuf, uintptr_t count,
                                 yaksi_type_s * type, yaksi_info_s * info, yaksa_op_t op,
                                 yaksi_request_s * request, yaksuri_request_s * reqpriv,
                                 yaksuri_subreq_s ** subreq_ptr)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;
    bool aligned = buf_is_aligned(inbuf, type);

    /* Fast path for REPLACE with aligned inbuf on the same device or different
     * devices with P2P support. Note that IPC mapping requires P2P support, thus
     * we don't handle the IPC case explicitly. */
    if (op == YAKSA_OP__REPLACE &&
        (request->backend.inattr.device == request->backend.outattr.device ||
         check_p2p_comm(id, reqpriv->request->backend.inattr.device,
                        reqpriv->request->backend.outattr.device)) && aligned) {
        rc = singlechunk_unpack(id, request->backend.inattr.device, inbuf, outbuf, count,
                                type, info, op, request, subreq_ptr);
    }
    /* Fast path for other reduce operations with aligned buffer on the same device */
    else if (request->backend.inattr.device == request->backend.outattr.device && aligned) {
        rc = singlechunk_unpack(id, request->backend.inattr.device, inbuf, outbuf, count,
                                type, info, op, request, subreq_ptr);
    }
    /* Slow paths */
    else if (request->backend.inattr.device == request->backend.outattr.device) {
        rc = set_multichunk_subreq(inbuf, outbuf, count, type, op, subreq_ptr);
        YAKSU_ERR_CHECK(rc, fn_fail);

        (*subreq_ptr)->u.multiple.acquire = unpack_d2d_unaligned_acquire;
        (*subreq_ptr)->u.multiple.release = simple_release;
    } else {
        bool is_enabled = check_p2p_comm(id, reqpriv->request->backend.inattr.device,
                                         reqpriv->request->backend.outattr.device);
        rc = set_multichunk_subreq(inbuf, outbuf, count, type, op, subreq_ptr);
        YAKSU_ERR_CHECK(rc, fn_fail);

        if (is_enabled) {
            (*subreq_ptr)->u.multiple.acquire = unpack_d2d_p2p_acquire;
        } else {
            (*subreq_ptr)->u.multiple.acquire = unpack_d2d_nop2p_acquire;
        }
        (*subreq_ptr)->u.multiple.release = simple_release;

    }

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int set_subreq_unpack_d2m(const void *inbuf, void *outbuf, uintptr_t count,
                                 yaksi_type_s * type, yaksi_info_s * info, yaksa_op_t op,
                                 yaksi_request_s * request, yaksuri_request_s * reqpriv,
                                 yaksuri_subreq_s ** subreq_ptr)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;

    /* Fast path for all reduce operations with aligned inbuf to host or the same device */
    if ((request->backend.outattr.device == -1 ||
         request->backend.inattr.device == request->backend.outattr.device) &&
        buf_is_aligned(inbuf, type)) {
        rc = singlechunk_unpack(id, request->backend.inattr.device, inbuf, outbuf, count,
                                type, info, op, request, subreq_ptr);
    }
    /* Slow path */
    else {
        rc = set_multichunk_subreq(inbuf, outbuf, count, type, op, subreq_ptr);
        YAKSU_ERR_CHECK(rc, fn_fail);

        (*subreq_ptr)->u.multiple.acquire = unpack_d2h_acquire;
        (*subreq_ptr)->u.multiple.release = unpack_d2h_release;
    }

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int set_subreq_unpack_d2rh(const void *inbuf, void *outbuf, uintptr_t count,
                                  yaksi_type_s * type, yaksi_info_s * info, yaksa_op_t op,
                                  yaksi_request_s * request, yaksuri_request_s * reqpriv,
                                  yaksuri_subreq_s ** subreq_ptr)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;
    uintptr_t threshold = yaksuri_global.gpudriver[id].hooks->get_iov_unpack_threshold(info);

    /* Fast path for REPLACE with contig type or noncontig type with large contig chunk */
    if (op == YAKSA_OP__REPLACE && (type->is_contig || type->size / type->num_contig >= threshold)) {
        rc = singlechunk_unpack(id, request->backend.inattr.device, inbuf, outbuf, count,
                                type, info, op, request, subreq_ptr);
    }
    /* Slow path */
    else {
        rc = set_multichunk_subreq(inbuf, outbuf, count, type, op, subreq_ptr);
        YAKSU_ERR_CHECK(rc, fn_fail);

        (*subreq_ptr)->u.multiple.acquire = unpack_d2h_acquire;
        (*subreq_ptr)->u.multiple.release = unpack_d2h_release;
    }

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int set_subreq_unpack_from_device(const void *inbuf, void *outbuf, uintptr_t count,
                                         yaksi_type_s * type, yaksi_info_s * info, yaksa_op_t op,
                                         yaksi_request_s * request, yaksuri_request_s * reqpriv,
                                         yaksuri_subreq_s ** subreq_ptr)
{
    int rc = YAKSA_SUCCESS;

    switch (request->backend.outattr.type) {
        case YAKSUR_PTR_TYPE__GPU:
            rc = set_subreq_unpack_d2d(inbuf, outbuf, count, type, info, op, request, reqpriv,
                                       subreq_ptr);
            break;
        case YAKSUR_PTR_TYPE__MANAGED:
            rc = set_subreq_unpack_d2m(inbuf, outbuf, count, type, info, op, request, reqpriv,
                                       subreq_ptr);
            break;
        case YAKSUR_PTR_TYPE__REGISTERED_HOST:
            rc = set_subreq_unpack_d2rh(inbuf, outbuf, count, type, info, op, request, reqpriv,
                                        subreq_ptr);
            break;
        case YAKSUR_PTR_TYPE__UNREGISTERED_HOST:
        default:
            rc = set_multichunk_subreq(inbuf, outbuf, count, type, op, subreq_ptr);
            YAKSU_ERR_CHECK(rc, fn_fail);

            (*subreq_ptr)->u.multiple.acquire = unpack_d2h_acquire;
            (*subreq_ptr)->u.multiple.release = unpack_d2h_release;
    }

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int set_subreq_unpack_from_managed(const void *inbuf, void *outbuf, uintptr_t count,
                                          yaksi_type_s * type, yaksi_info_s * info, yaksa_op_t op,
                                          yaksi_request_s * request, yaksuri_request_s * reqpriv,
                                          yaksuri_subreq_s ** subreq_ptr)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;

    /* Fast path for all reduce operations with aligned inbuf from host or the same device. */
    if ((request->backend.inattr.device == -1 ||
         request->backend.inattr.device == request->backend.outattr.device) &&
        buf_is_aligned(inbuf, type)) {
        rc = singlechunk_unpack(id, request->backend.outattr.device, inbuf, outbuf, count,
                                type, info, op, request, subreq_ptr);
    }
    /* Slow path */
    else {
        rc = set_multichunk_subreq(inbuf, outbuf, count, type, op, subreq_ptr);
        YAKSU_ERR_CHECK(rc, fn_fail);

        (*subreq_ptr)->u.multiple.acquire = unpack_urh2d_acquire;
        (*subreq_ptr)->u.multiple.release = simple_release;
    }

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

static int set_subreq_unpack_from_rhost(const void *inbuf, void *outbuf, uintptr_t count,
                                        yaksi_type_s * type, yaksi_info_s * info, yaksa_op_t op,
                                        yaksi_request_s * request, yaksuri_request_s * reqpriv,
                                        yaksuri_subreq_s ** subreq_ptr)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;
    uintptr_t threshold = yaksuri_global.gpudriver[id].hooks->get_iov_unpack_threshold(info);

    /* Fast path for REPLACE with contig type or noncontig type with large contig chunk */
    if (op == YAKSA_OP__REPLACE && (type->is_contig || type->size / type->num_contig >= threshold)) {
        rc = singlechunk_unpack(id, request->backend.outattr.device, inbuf, outbuf, count,
                                type, info, op, request, subreq_ptr);
    }
    /* Slow path */
    else {
        rc = set_multichunk_subreq(inbuf, outbuf, count, type, op, subreq_ptr);
        YAKSU_ERR_CHECK(rc, fn_fail);

        (*subreq_ptr)->u.multiple.acquire = unpack_rh2d_acquire;
        (*subreq_ptr)->u.multiple.release = simple_release;
    }

  fn_exit:
    return rc;
  fn_fail:
    goto fn_exit;
}

int yaksuri_progress_enqueue(const void *inbuf, void *outbuf, uintptr_t count, yaksi_type_s * type,
                             yaksi_info_s * info, yaksa_op_t op, yaksi_request_s * request)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_subreq_s *subreq = NULL;
    yaksuri_request_s *reqpriv = (yaksuri_request_s *) request->backend.priv;
    yaksuri_gpudriver_id_e id = reqpriv->gpudriver_id;

    assert(yaksuri_global.gpudriver[id].hooks);
    assert(request->backend.inattr.type == YAKSUR_PTR_TYPE__GPU ||
           request->backend.outattr.type == YAKSUR_PTR_TYPE__GPU);

    reqpriv->info = info;

    /* if the GPU reqpriv cannot support this type, return */
    bool is_supported;
    rc = yaksuri_global.gpudriver[id].hooks->pup_is_supported(type, op, &is_supported);
    YAKSU_ERR_CHECK(rc, fn_fail);

    if (!is_supported) {
        rc = YAKSA_ERR__NOT_SUPPORTED;
        goto fn_exit;
    }

    if (reqpriv->optype == YAKSURI_OPTYPE__PACK) {
        switch (request->backend.inattr.type) {
            case YAKSUR_PTR_TYPE__GPU:
                rc = set_subreq_pack_from_device(inbuf, outbuf, count, type, info, op, request,
                                                 reqpriv, &subreq);
                YAKSU_ERR_CHECK(rc, fn_fail);
                break;
            case YAKSUR_PTR_TYPE__MANAGED:
                rc = set_subreq_pack_from_managed(inbuf, outbuf, count, type, info, op, request,
                                                  reqpriv, &subreq);
                YAKSU_ERR_CHECK(rc, fn_fail);
                break;
            case YAKSUR_PTR_TYPE__REGISTERED_HOST:
                rc = set_subreq_pack_from_rhost(inbuf, outbuf, count, type, info, op, request,
                                                reqpriv, &subreq);
                YAKSU_ERR_CHECK(rc, fn_fail);
                break;
            case YAKSUR_PTR_TYPE__UNREGISTERED_HOST:
            default:
                rc = set_multichunk_subreq(inbuf, outbuf, count, type, op, &subreq);
                YAKSU_ERR_CHECK(rc, fn_fail);

                subreq->u.multiple.acquire = pack_h2d_acquire;
                subreq->u.multiple.release = simple_release;
                break;
        }
    } else {
        switch (request->backend.inattr.type) {
            case YAKSUR_PTR_TYPE__GPU:
                rc = set_subreq_unpack_from_device(inbuf, outbuf, count, type, info, op, request,
                                                   reqpriv, &subreq);
                YAKSU_ERR_CHECK(rc, fn_fail);
                break;
            case YAKSUR_PTR_TYPE__MANAGED:
                rc = set_subreq_unpack_from_managed(inbuf, outbuf, count, type, info, op, request,
                                                    reqpriv, &subreq);
                YAKSU_ERR_CHECK(rc, fn_fail);
                break;
            case YAKSUR_PTR_TYPE__REGISTERED_HOST:
                rc = set_subreq_unpack_from_rhost(inbuf, outbuf, count, type, info, op, request,
                                                  reqpriv, &subreq);
                YAKSU_ERR_CHECK(rc, fn_fail);
                break;
            case YAKSUR_PTR_TYPE__UNREGISTERED_HOST:
            default:
                rc = set_multichunk_subreq(inbuf, outbuf, count, type, op, &subreq);
                YAKSU_ERR_CHECK(rc, fn_fail);

                subreq->u.multiple.acquire = unpack_urh2d_acquire;
                subreq->u.multiple.release = simple_release;
                break;
        }
    }

    /* If pack/unpack is already completed, no subreq is returned. Thus, skip enqueue. */
    if (!subreq) {
        goto fn_exit;
    }

    pthread_mutex_lock(&progress_mutex);
    DL_APPEND(reqpriv->subreqs, subreq);

    /* if the request is not in our pending list, add it */
    yaksuri_request_s *req;
    HASH_FIND_PTR(pending_reqs, &request, req);
    if (req == NULL) {
        HASH_ADD_PTR(pending_reqs, request, reqpriv);
        yaksu_atomic_incr(&request->cc);
    }
    pthread_mutex_unlock(&progress_mutex);

    rc = yaksuri_progress_poke();
    YAKSU_ERR_CHECK(rc, fn_fail);

  fn_exit:
    return rc;
  fn_fail:
    free(subreq);
    goto fn_exit;
}

int yaksuri_progress_poke(void)
{
    int rc = YAKSA_SUCCESS;
    yaksuri_gpudriver_id_e id = YAKSURI_GPUDRIVER_ID__UNSET;

    /* A progress poke is in two steps.  In the first step, we check
     * for event completions, finish any post-processing and retire
     * any temporary resources.  In the second steps, we issue out any
     * pending operations. */

    pthread_mutex_lock(&progress_mutex);

    yaksi_type_s *byte_type;
    rc = yaksi_type_get(YAKSA_TYPE__BYTE, &byte_type);
    YAKSU_ERR_CHECK(rc, fn_fail);

    /**********************************************************************/
    /* Step 1: Check for completions */
    /**********************************************************************/
    yaksuri_request_s *reqpriv, *tmp;
    HASH_ITER(hh, pending_reqs, reqpriv, tmp) {
        id = reqpriv->gpudriver_id;
        assert(reqpriv->subreqs);

        yaksuri_subreq_s *subreq, *tmp2;
        DL_FOREACH_SAFE(reqpriv->subreqs, subreq, tmp2) {
            if (subreq->kind == YAKSURI_SUBREQ_KIND__SINGLE_CHUNK) {
                int completed;
                rc = event_query(id, subreq->u.single.event, &completed);
                YAKSU_ERR_CHECK(rc, fn_fail);

                if (!completed)
                    continue;

                DL_DELETE(reqpriv->subreqs, subreq);
                free(subreq);
                if (reqpriv->subreqs == NULL) {
                    HASH_DEL(pending_reqs, reqpriv);
                    yaksu_atomic_decr(&reqpriv->request->cc);
                }
            } else {
                yaksuri_subreq_chunk_s *chunk, *tmp3;
                DL_FOREACH_SAFE(subreq->u.multiple.chunks, chunk, tmp3) {
                    int completed;
                    rc = event_query(id, chunk->event, &completed);
                    YAKSU_ERR_CHECK(rc, fn_fail);

                    if (!completed)
                        continue;

                    rc = subreq->u.multiple.release(reqpriv, subreq, chunk);
                    YAKSU_ERR_CHECK(rc, fn_fail);
                }
            }
        }
    }

    /**********************************************************************/
    /* Step 2: Issue new operations */
    /**********************************************************************/
    HASH_ITER(hh, pending_reqs, reqpriv, tmp) {
        id = reqpriv->gpudriver_id;
        assert(reqpriv->subreqs);

        yaksuri_subreq_s *subreq, *tmp2;
        DL_FOREACH_SAFE(reqpriv->subreqs, subreq, tmp2) {
            if (subreq->kind == YAKSURI_SUBREQ_KIND__SINGLE_CHUNK)
                continue;

            while (subreq->u.multiple.issued_count < subreq->u.multiple.count) {
                yaksuri_subreq_chunk_s *chunk;

                rc = subreq->u.multiple.acquire(reqpriv, subreq, &chunk);
                YAKSU_ERR_CHECK(rc, fn_fail);

                if (chunk == NULL)
                    goto flush;

                subreq->u.multiple.issued_count += chunk->count;
            }
        }
    }

  flush:
    /* if we issued any operations, call flush_all, so the driver
     * layer can flush the kernels. */
    if (id != YAKSURI_GPUDRIVER_ID__UNSET) {
        rc = yaksuri_global.gpudriver[id].hooks->flush_all();
        YAKSU_ERR_CHECK(rc, fn_fail);
    }

  fn_exit:
    pthread_mutex_unlock(&progress_mutex);
    return rc;
  fn_fail:
    goto fn_exit;
}
